@@ -1,6 +1,6 @@
 #!/usr/bin/perl -w
 
-# Lintian reporting harness -- Run lintian against a archive mirror
+# Lintian reporting harness -- Run lintian against an archive mirror
 #
 # Copyright (C) 2015 Niels Thykier
 #
@@ -23,8 +23,9 @@
 # Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 # MA 02110-1301, USA.
 
-use strict;
+use v5.20;
 use warnings;
+use utf8;
 use autodie;
 
 use constant BACKLOG_PROCESSING_TIME_LIMIT => 4 * 3600; # 4hours
@@ -37,7 +38,8 @@ use List::MoreUtils qw(first_index);
 use POSIX qw(strftime);
 
 use Lintian::Command qw(safe_qx);
-use Lintian::Util qw(find_backlog load_state_cache save_state_cache untaint);
+use Lintian::Reporting::Util
+  qw(find_backlog load_state_cache save_state_cache);
 
 my (@LINTIAN_CMD, $LINTIAN_VERSION);
 
@@ -47,19 +49,14 @@ my @REQUIRED_PARAMETERS = qw(
   schedule-limit-groups
   state-dir
 );
-my %OPT = (
-    'use-permanent-lab' => 'guess',
-    'lintian-frontend'  => 'lintian',
-);
+my %OPT = ('lintian-frontend'  => 'lintian',);
 my %OPT_HASH = (
     'schedule-chunk-size=i'   => \$OPT{'schedule-chunk-size'},
     'schedule-limit-groups=i' => \$OPT{'schedule-limit-groups'},
     'state-dir=s'             => \$OPT{'state-dir'},
     'lintian-frontend=s'      => \$OPT{'lintian-frontend'},
     'lintian-log-dir=s'       => \$OPT{'lintian-log-dir'},
-    'lintian-lab=s'           => \$OPT{'lintian-lab'},
     'lintian-scratch-space=s' => \$OPT{'lintian-scratch-space'},
-    'use-permanent-lab!'      => \$OPT{'use-permanent-lab'},
     'help|h'                  => \&usage,
 );
 
@@ -76,6 +73,17 @@ sub main {
 
 ### END OF SCRIPT -- below are helper subroutines ###
 
+=item untaint(VALUE)
+
+Untaint VALUE
+
+=cut
+
+sub untaint {
+    return $_[0] = $1 if $_[0] =~ m/^(.*)$/;
+    return;
+}
+
 sub check_parameters {
     for my $parameter (@REQUIRED_PARAMETERS) {
         if (not defined($OPT{$parameter})) {
@@ -90,18 +98,7 @@ sub check_parameters {
     }
     die("The argument for --schedule-limit-groups must be an > 0\n")
       if $OPT{'schedule-limit-groups'} < 1;
-    if ($OPT{'use-permanent-lab'} eq 'guess') {
-        $OPT{'use-permanent-lab'} = 0;
-        $OPT{'use-permanent-lab'} = 1 if $OPT{'lintian-lab'};
-    } elsif ($OPT{'use-permanent-lab'} and not $OPT{'lintian-lab'}) {
-        die(    'If --use-permanent-lab is given, then'
-              . " --lintian-lab must be given too\n");
-    } elsif (not $OPT{'use-permanent-lab'} and $OPT{'lintian-lab'}) {
-        warn(   'Ignoring --lintian-lab when explicit'
-              . " --no-use-permanent-lab is given\n");
-        warn("Perhaps you wanted --lintian-scratch-space <PATH> instead?\n");
-        delete($OPT{'lintian-lab'});
-    }
+
     return;
 }
 
@@ -109,7 +106,7 @@ sub prepare_lintian_environment_and_cmdline {
     my $frontend = 'lintian';
     my $eoa_marker_index = first_index { $_ eq '--' } @ARGV;
     my $logs_dir = $OPT{'lintian-log-dir'};
-    my @overridable_args = (qw(-EIL +pedantic --show-overrides));
+    my @overridable_args = (qw(-EL +>=classification --show-overrides));
     my @args = (
         qw(--verbose), # We rely on this for filtering the log
         qw(--exp-output=format=fullewi --packages-from-file -),
@@ -117,9 +114,7 @@ sub prepare_lintian_environment_and_cmdline {
         "+${logs_dir}/lintian-perf.log",
     );
     $frontend = $OPT{'lintian-frontend'} if ($OPT{'lintian-frontend'});
-    if ($OPT{'use-permanent-lab'}) {
-        push(@args, '--lab', $OPT{'lintian-lab'});
-    }
+
     if ($eoa_marker_index > -1) {
         # Move known "non-parameters" and the "--" behind our arguments.
         # It is a misfeature, but at least it does not break
@@ -160,10 +155,6 @@ sub harness_lintian {
 
     @worklist = find_backlog($LINTIAN_VERSION, $state);
 
-    # Sanity check for incremental mode.
-    die("Old Lintian log file $lintian_log not found!\n")
-      if ($OPT{'incremental-mode'} and not -f $lintian_log);
-
     # Always update the log if it exists, as we may have removed
     # some entries.
     if (-f $lintian_log) {
@@ -217,10 +208,12 @@ sub process_worklist {
     log_msg('Command line used: ' . join(q{ }, @LINTIAN_CMD));
     while (@worklist) {
         my $len = scalar @worklist;
-        my (@work_splice, @completed, %processed);
+        my (@work_splice, @completed, %processed, %errors);
         my ($lintpipe, $lint_stdin, $status_fd, $lint_status_out);
+        my $got_alarm = 0;
 
-        if (time() >= $start_time + BACKLOG_PROCESSING_TIME_LIMIT) {
+        # Bail if there is less than 5 minutes left
+        if (time() >= $start_time + BACKLOG_PROCESSING_TIME_LIMIT - 300) {
             log_msg('No more time for processing backlogs');
             $exit_code = 2;
             last;
@@ -294,26 +287,73 @@ sub process_worklist {
                 my $path = $members->{$member_id}{'path'};
                 print {$lintpipe} "$path\n";
             }
-            #print {$lintpipe} "!query: GROUP:$group_id\n";
         }
         close($lintpipe);
 
-        # Listen to status updates from lintian
-        while (my $line = <$status_fd>) {
-            chomp($line);
-            if ($line =~ s/^complete ([^ ]+) \(([^\)]+)\)$//) {
-                my ($group_id, $runtime) = ($1, $2);
-                push(@completed, $group_id);
-                $processed{$group_id} = 1;
-                log_msg("  [lintian] processed $group_id"
-                      . " successfully (time: $runtime)");
-            } elsif ($line =~ s/^error ([^ ]+) \(([^\)]+)\)$//) {
-                my ($group_id, $runtime) = ($1, $2);
-                log_msg(
-                    "  [lintian] error processing $group_id (time: $runtime)");
-                $processed{$group_id} = 1;
+        eval {
+            my $time_limit
+              = $start_time + BACKLOG_PROCESSING_TIME_LIMIT - time();
+            my $count = 0;
+            my $signalled_lintian = 0;
+            my $sig_handler = sub {
+                my ($signal_name) = @_;
+                $signalled_lintian = 1;
+                $count++;
+                if ($signal_name eq 'ALRM') {
+                    $got_alarm = 1 if $got_alarm >= 0;
+                } else {
+                    $got_alarm = -1;
+                }
+                if ($count < 3) {
+                    log_msg("Received SIG${signal_name}, "
+                          . "sending SIGTERM to $pid [${count}/3]");
+                    kill('TERM', $pid);
+                    if ($signal_name eq 'ALRM') {
+                        log_msg(
+                            'Scheduling another alarm in 5 minutes from now...'
+                        );
+                        alarm(300);
+                    }
+                } else {
+                    log_msg("Received SIG${signal_name} as the third one, "
+                          . "sending SIGKILL to $pid");
+                    log_msg('You may have to clean up some '
+                          . 'temporary directories manually');
+                    kill('KILL', $pid);
+                }
+            };
+            local $SIG{'TERM'} = $sig_handler;
+            local $SIG{'INT'} = $sig_handler;
+            local $SIG{'ALRM'} = $sig_handler;
+
+            alarm($time_limit);
+
+            # Listen to status updates from lintian
+            while (my $line = <$status_fd>) {
+                chomp($line);
+                if ($line =~ m/^complete ([^ ]+) \(([^\)]+)\)$/) {
+                    my ($group_id, $runtime) = ($1, $2);
+                    push(@completed, $group_id);
+                    $processed{$group_id} = 1;
+                    log_msg("  [lintian] processed $group_id"
+                          . " successfully (time: $runtime)");
+                } elsif ($line =~ m/^error ([^ ]+) \(([^\)]+)\)$/) {
+                    my ($group_id, $runtime) = ($1, $2);
+                    log_msg("  [lintian] error processing $group_id "
+                          . "(time: $runtime)");
+                    $processed{$group_id} = 1;
+                    # We ignore errors if we sent lintian a signal to avoid
+                    # *some* false-positives.
+                    $errors{$group_id} = 1 if not $signalled_lintian;
+                } elsif ($line =~ m/^ack-signal (SIG\S+)$/) {
+                    my $signal = $1;
+                    log_msg(
+                        "Signal $signal acknowledged: disabled timed alarms");
+                    alarm(0);
+                }
             }
-        }
+            alarm(0);
+        };
         close($status_fd);
 
         # Wait for lintian to terminate
@@ -325,17 +365,36 @@ sub process_worklist {
             my $sig = $? & 0xff;
             if ($res != 1 and $res != 0) {
                 log_msg("warning: executing lintian returned $res");
+                if ($got_alarm) {
+                    # Ideally, lintian would always die by the signal
+                    # but some times it catches it and terminates
+                    # "normally"
+                    log_msg('Stopped by a signal or time out');
+                    log_msg(' - skipping the rest of the worklist');
+                    @worklist = ();
+                }
             } elsif ($sig) {
                 log_msg("Lintian terminated by signal: $sig");
                 # If someone is sending us signals (e.g. SIGINT/Ctrl-C)
                 # don't start the next round.
                 log_msg(' - skipping the rest of the worklist');
                 @worklist = ();
-                $exit_code = 3;
+            }
+            if ($got_alarm) {
+                if ($got_alarm == 1) {
+                    # Lintian was (presumably) killed due to a
+                    # time-out from this process
+                    $exit_code = 2;
+                } else {
+                    # Lintian was killed by another signal; notify
+                    # harness that it should skip the rest as well.
+                    $exit_code = 3;
+                }
             }
         } else {
             log_msg('Lintian finished successfully');
         }
+        log_msg('Updating the lintian log used for reporting');
         my $filter = generate_log_filter($state, \%processed);
         seek($nfd, 0, SEEK_END);
         update_lintian_log($filter, $nfd, $new_lintian_log);
@@ -352,6 +411,32 @@ sub process_worklist {
             $group_data = $state->{'groups'}{$group_id};
             $group_data->{'last-processed-by'} = $LINTIAN_VERSION;
             delete($group_data->{'out-of-date'});
+            # Always clear the error counter after a successful run.
+            delete($group_data->{'processing-errors'});
+            delete($group_data->{'last-error-by'});
+        }
+        for my $group_id (sort(keys(%errors))) {
+            my $group_data;
+            # In theory, they can disappear - in practise, that requires
+            # an external call to (e.g.) dplint reporting-sync-state.
+            next if not exists($state->{'groups'}{$group_id});
+            $group_data = $state->{'groups'}{$group_id};
+            if ($errors{$group_id}) {
+                if (not exists($group_data->{'last-error-by'})
+                    or $group_data->{'last-error-by'} ne $LINTIAN_VERSION) {
+                    # If it is a new lintian version then (re)set the counter
+                    # to 1.  Case also triggers for the very first issue.
+                    $group_data->{'processing-errors'} = 1;
+                } else {
+                    # Repeated error with the same version
+                    ++$group_data->{'processing-errors'};
+                }
+                # Set the "last-error-by" flag so we can clear the
+                # error if there is a new version of lintian.
+                $group_data->{'last-error-by'} = $LINTIAN_VERSION;
+            } else {
+                delete($group_data->{'processing-errors'});
+            }
         }
         save_state_cache($OPT{'state-dir'}, $state);
         last if $exit_code;
@@ -388,7 +473,7 @@ sub update_lintian_log {
                 m/^N: [ ] Processing [ ] (binary|udeb|source) [ ]
                        package [ ] (\S+) [ ] \(version [ ] (\S+), [ ]
                        arch [ ] (\S+)\)[ ]\.\.\./oxsm
-              ) {
+            ) {
                 my ($type, $pkg, $ver, $arch) = ($1,$2, $3, $4);
                 my $k = "$type:$pkg/$ver";
                 $k .= "/$arch" if $type ne 'source';
@@ -440,15 +525,11 @@ Usage: $me <args> -- <extra lintian args>
 
   --lintian-frontend PROG     Use PROG as frontend for lintian (defaults to "lintian")
   --lintian-log-dir DIR       Path to the harness log dir. [!]
-  --lintian-lab DIR           Use DIR as permanent lab (implies --use-permanent-lab).
-                              Ignored with --no-use-permanent-lab.
   --lintian-scratch-space DIR Use DIR for temporary files (notably temp labs)
   --schedule-chunk-size N     Run at most N groups in a given lintian run.
   --schedule-limit-groups N   Schedule at most N groups in total. [!]
   --state-dir DIR             Directory containing the state cache (must be
                               writable). [!]
-  --[no-]use-permanent-lab    Whether to use a permanent lab for lintian instead of
-                              throw away labs.
 
 Arguments marked with [!] are required for a successful run.
 EOF
