@@ -18,27 +18,12 @@
 
 package Lintian::Output;
 
-use strict;
+use v5.20;
 use warnings;
+use utf8;
 
-use v5.8.0; # for PerlIO
-use parent qw(Class::Accessor::Fast);
-# The limit is including the "use --foo to see all tags".
-use constant DEFAULT_INTERACTIVE_TAG_LIMIT => 4;
-
-use Exporter qw(import);
-
-# Force export as soon as possible, since some of the modules we load also
-# depend on us and the sequencing can cause things not to be exported
-# otherwise.
-our (%EXPORT_TAGS, @EXPORT_OK);
-
-BEGIN {
-    %EXPORT_TAGS = (
-        messages => [qw(msg v_msg warning debug_msg delimiter perf_log)],
-        util => [qw(_global_or_object)]);
-    @EXPORT_OK = (@{$EXPORT_TAGS{messages}},@{$EXPORT_TAGS{util}},'string');
-}
+use Moo::Role;
+use namespace::clean;
 
 =head1 NAME
 
@@ -71,33 +56,15 @@ Lintian::Output is used for all interaction between lintian and the user.
 It is designed to be easily extensible via subclassing.
 
 To simplify usage in the most common cases, many Lintian::Output methods
-can be used as class methods and will therefor automatically use the object
+can be used as class methods and will therefore automatically use the object
 $Lintian::Output::GLOBAL unless their first argument C<isa('Lintian::Output')>.
 
-=cut
+=head1 ATTRIBUTES
 
-use Lintian::Tags ();
-
-# support for ANSI color output via colored()
-use Term::ANSIColor ();
-
-=head1 ACCESSORS
-
-The following fields define the behaviours of Lintian::Output.
+The following fields impact the behavior of Lintian::Output.
 
 =over 4
 
-=item verbosity_level
-
-Determine how verbose the output should be.  "0" is the default value
-(tags and msg only), "-1" is quiet (tags only) and "1" is verbose
-(tags, msg and v_msg).
-
-=item debug
-
-If set to a positive integer, will enable all debug messages issued with
-a level lower or equal to its value.
-
 =item color
 
 Can take the values "never", "always", "auto" or "html".
@@ -109,6 +76,23 @@ color only if the output is going to a terminal.
 "html" will output HTML <span> tags with a color style attribute (instead
 of ANSI color escape sequences).
 
+=item colors
+
+=item debug
+
+If set to a positive integer, will enable all debug messages issued with
+a level lower or equal to its value.
+
+=item issuedtags
+
+Hash containing the names of tags which have been issued.
+
+=item perf_debug
+
+=item perf_log_fd
+
+=item proc_id2tag_count
+
 =item stdout
 
 I/O handle to use for output of messages and tags.  Defaults to C<\*STDOUT>.
@@ -121,54 +105,46 @@ I/O handle to use for warnings.  Defaults to C<\*STDERR>.
 
 Whether to show the description of a tag when printing it.
 
-=item issuedtags
-
-Hash containing the names of tags which have been issued.
+=item tty_hyperlinks
 
 =item tag_display_limit
 
 Get/Set the number of times a tag is emitted per processable.
 
+=item verbosity_level
+
+Determine how verbose the output should be.  "0" is the default value
+(tags and msg only), "-1" is quiet (tags only) and "1" is verbose
+(tags, msg and v_msg).
+
 =back
 
 =cut
 
-Lintian::Output->mk_accessors(
-    qw(verbosity_level debug color colors stdout
-      stderr perf_log_fd perf_debug showdescription
-      issuedtags tag_display_limit)
-);
+has color => (is => 'rw', default => 'never');
+has colors => (
+    is => 'rw',
+    default => sub {
+        {
+            'E' => 'red',
+            'W' => 'yellow',
+            'I' => 'cyan',
+            'P' => 'green',
+            'C' => 'blue',
+        }
+    });
+has issuedtags => (is => 'rw', default => sub { {} });
+has perf_debug => (is => 'rw', default => 0);
+has perf_log_fd => (is => 'rw', default => sub { \*STDOUT });
+has proc_id2tag_count => (is => 'rw', default => sub { {} });
+has stderr => (is => 'rw', default => sub { \*STDERR });
+has stdout => (is => 'rw', default => sub { \*STDOUT });
+has tag_display_limit => (is => 'rw', default => 4);
+has tty_hyperlinks => (is => 'rw', default => 0);
+has verbosity_level => (is => 'rw', default => 0);
 
-# for the non-OO interface
-my %default_colors = (
-    'E' => 'red',
-    'W' => 'yellow',
-    'I' => 'cyan',
-    'P' => 'green',
-    'C' => 'blue',
-);
-
-our $GLOBAL = Lintian::Output->new;
-
-sub new {
-    my ($class, %options) = @_;
-    my $self = {%options};
-
-    bless($self, $class);
-
-    $self->stdout(\*STDOUT);
-    $self->stderr(\*STDERR);
-    $self->perf_log_fd(\*STDOUT);
-    $self->colors({%default_colors});
-    $self->issuedtags({});
-    $self->{'proc_id2tag_count'} = {};
-
-    # Set defaults to avoid "uninitialized" warnings
-    $self->verbosity_level(0);
-    $self->perf_debug(0);
-    $self->color('never');
-    return $self;
-}
+has debug => (is => 'rw', default => sub { {} });
+has showdescription => (is => 'rw', default => sub { {} });
 
 =head1 CLASS/INSTANCE METHODS
 
@@ -198,7 +174,7 @@ with 'N: '.  Will do nothing unless debug is set to a positive integer
 =cut
 
 sub msg {
-    my ($self, @args) = _global_or_object(@_);
+    my ($self, @args) = @_;
 
     return if $self->verbosity_level < 0;
     $self->_message(@args);
@@ -206,7 +182,7 @@ sub msg {
 }
 
 sub v_msg {
-    my ($self, @args) = _global_or_object(@_);
+    my ($self, @args) = @_;
 
     return unless $self->verbosity_level > 0;
     $self->_message(@args);
@@ -214,7 +190,7 @@ sub v_msg {
 }
 
 sub debug_msg {
-    my ($self, $level, @args) = _global_or_object(@_);
+    my ($self, $level, @args) = @_;
 
     return unless $self->debug && ($self->debug >= $level);
 
@@ -230,7 +206,7 @@ prefixed with 'warning: '.
 =cut
 
 sub warning {
-    my ($self, @args) = _global_or_object(@_);
+    my ($self, @args) = @_;
 
     return if $self->verbosity_level < 0;
     $self->_warning(@args);
@@ -249,7 +225,7 @@ positive integer.
 =cut
 
 sub perf_log {
-    my ($self, @args) = _global_or_object(@_);
+    my ($self, @args) = @_;
 
     return unless $self->perf_debug;
 
@@ -268,7 +244,7 @@ with one of the methods above, e.g.
 =cut
 
 sub delimiter {
-    my ($self) = _global_or_object(@_);
+    my ($self) = @_;
 
     return $self->_delimiter;
 }
@@ -281,7 +257,7 @@ indicating whether the tag had previously been issued by the object.
 =cut
 
 sub issued_tag {
-    my ($self, $tag_name) = _global_or_object(@_);
+    my ($self, $tag_name) = @_;
 
     return $self->issuedtags->{$tag_name}++ ? 1 : 0;
 }
@@ -293,7 +269,7 @@ TODO: Is this part of the public interface?
 =cut
 
 sub string {
-    my ($self, $lead, @args) = _global_or_object(@_);
+    my ($self, $lead, @args) = @_;
 
     my $output = '';
     if (@args) {
@@ -311,133 +287,6 @@ sub string {
 
 =back
 
-=head1 INSTANCE METHODS FOR CONTEXT-AWARE OUTPUT
-
-The following methods are designed to be called at specific points
-during program execution and require very specific arguments.  They
-can only be called as instance methods.
-
-=over 4
-
-=item C<print_tag($pkg_info, $tag_info, $extra, $override)>
-
-Print a tag.  The first two arguments are hash reference with the
-information about the package and the tag, $extra is the extra
-information for the tag (if any) as an array reference, and $override
-is either undef if the tag is not overridden or the
-L<override|Lintian::Tag::Override> for this tag.  Called from
-Lintian::Tags::tag().
-
-=cut
-
-sub print_tag {
-    my ($self, $pkg_info, $tag_info, $information, $override) = @_;
-    $information = ' ' . $self->_quote_print($information)
-      if $information ne '';
-    my $code = $tag_info->code;
-    my $tag_color = $self->{colors}{$code};
-    my $fpkg_info = $self->_format_pkg_info($pkg_info, $tag_info, $override);
-    my $tag_name = $tag_info->tag;
-    my $limit = $self->tag_display_limit;
-    my $tag;
-
-    # Limit the output so people do not drown in tags.  Some tags are
-    # insanely noisy (hi static-library-has-unneeded-section)
-    if ($limit) {
-        my $proc_id = $pkg_info->{'processable'}->identifier;
-        my $emitted_count
-          = $self->{'proc_id2tag_count'}{$proc_id}{$tag_name}++;
-        return if $emitted_count >= $limit;
-        my $msg
-          = ' ... use --no-tag-display-limit to see all (or pipe to a file/program)';
-        $information = $self->_quote_print($msg)
-          if $emitted_count >= $limit-1;
-    }
-    if ($self->_do_color) {
-        if ($self->color eq 'html') {
-            my $escaped = $tag_name;
-            $escaped =~ s/&/&amp;/g;
-            $escaped =~ s/</&lt;/g;
-            $escaped =~ s/>/&gt;/g;
-            $tag .= qq(<span style="color: $tag_color">$escaped</span>);
-        } else {
-            $tag .= Term::ANSIColor::colored($tag_name, $tag_color);
-        }
-    } else {
-        $tag .= $tag_name;
-    }
-
-    if ($override && @{ $override->comments }) {
-        foreach my $c (@{ $override->comments }) {
-            $self->msg($self->_quote_print($c));
-        }
-    }
-
-    $self->_print('', $fpkg_info, "$tag$information");
-    if (not $self->issued_tag($tag_info->tag) and $self->showdescription) {
-        my $description;
-        if ($self->_do_color && $self->color eq 'html') {
-            $description = $tag_info->description('html', '   ');
-        } else {
-            $description = $tag_info->description('text', '   ');
-        }
-        $self->_print('', 'N', '');
-        $self->_print('', 'N', split("\n", $description));
-        $self->_print('', 'N', '');
-    }
-    return;
-}
-
-# Helper function to "print_tag" to decide the output format of the tag line.  Used by
-# the "FullEWI" subclass.
-#
-sub _format_pkg_info {
-    my ($self, $pkg_info, $tag_info, $override) = @_;
-    my $code = $tag_info->code;
-    $code = 'X' if $tag_info->experimental;
-    $code = 'O' if defined $override;
-    my $type = '';
-    $type = " $pkg_info->{type}" if $pkg_info->{type} ne 'binary';
-    return "$code: $pkg_info->{package}$type";
-}
-
-=item C<print_start_pkg($pkg_info)>
-
-Called before lintian starts to handle each package.  The version in
-Lintian::Output uses v_msg() for output.  Called from Tags::select_pkg().
-
-=cut
-
-sub print_start_pkg {
-    my ($self, $pkg_info) = @_;
-
-    my $object = 'package';
-    if ($pkg_info->{type} eq 'changes') {
-        $object = 'file';
-    }
-
-    $self->v_msg(
-        $self->delimiter,
-        join(q{ },
-            "Processing $pkg_info->{type} $object $pkg_info->{package}",
-            "(version $pkg_info->{version}, arch $pkg_info->{arch}) ..."));
-    return;
-}
-
-=item C<print_start_pkg($pkg_info)>
-
-Called after lintian is finished with a package.  The version in
-Lintian::Output does nothing.  Called from Lintian::Tags::file_start() and
-Lintian::Tags::file_end().
-
-=cut
-
-sub print_end_pkg {
-    return;
-}
-
-=back
-
 =head1 INSTANCE METHODS FOR SUBCLASSING
 
 The following methods are only intended for subclassing and are
@@ -530,46 +379,12 @@ sub _do_color {
             && -t $self->stdout));
 }
 
-=item C<_quote_print($string)>
-
-Called to quote a string.  By default it will replace all
-non-printables with "?".  Sub-classes can override it if
-they allow non-ascii printables etc.
-
-=cut
-
-sub _quote_print {
-    my ($self, $string) = @_;
-    $string =~ s/[^[:print:]]/?/go;
-    return $string;
-}
-
-=back
-
-=head1 CLASS METHODS
-
-=over 4
-
-=item C<_global_or_object(@args)>
-
-If $args[0] is an object which satisfies C<isa('Lintian::Output')>
-returns @args, otherwise returns C<($Lintian::Output::GLOBAL, @_)>.
-
-=back
-
-=cut
-
-sub _global_or_object {
-    if (ref($_[0]) and $_[0]->isa('Lintian::Output')) {
-        return @_;
-    } else {
-        return ($Lintian::Output::GLOBAL, @_);
-    }
-}
-
 1;
+
 __END__
 
+=back
+
 =head1 EXPORTS
 
 Lintian::Output exports nothing by default, but the following export
@@ -587,10 +402,6 @@ Exports all the methods in L<CLASS METHODS>
 
 =back
 
-=head1 AUTHOR
-
-Originally written by Frank Lichtenheld <djpig@debian.org> for Lintian.
-
 =head1 SEE ALSO
 
 lintian(1)
